;+
; NAME:
;       krprogressbar__define
;
; PURPOSE:
;
;       An object for creating a truly unique progress bar made up of image objects
;
; AUTHOR:
;
;       Kling Research And Software, Inc
;       Ronn Kling
;       7038 Westmoreland Dr.
;       Warrenton, VA 20187
;       ronn@rlkling.com
;
; CATEGORY:

;       Utilities
;
; CALLING SEQUENCE:
;
;       krProgressBar = obj_new("krprogressbar", lots_of_keywords)
;
; INPUTS:
;
;       parent: A widget identifier of the widget that will be the
;           group leader for this program. If not present then this is a
;           stand alone widget.
;
; KEYWORDS:
;
;      cancelButton = Set this keyword if a cancel button is desired.
;      endTime = The total time the widget should be on the display in autoUpDate
;           mode. The keyword applies only to AutoUpDate mode. Default is 10 seconds.
;      timerInterval = timer interval for auto update, non zero means this is active.
;      message = The text of the label below the progress bar. Default is "Working..."
;      title = The text of the top-level base title bar. Default is "krProgressBar"
;      xsize = The XSize of the progress bar in Device coordinates. Default is 150.
;      ysize = The YSize of the progress bar in Device coordinates. Default is 25.
;      backgroundImage = background image object, default is a blue bar
;      maskImage = mask image for sliding over the background image, default is a white image
;      maskFunction = function to call for the mask image opacity values, default is krLinear,
;					  but krstep is also built in
;      percentInterval = default percentage increment (10%)
;      displayPercent = if set then percentage is displayed
;      textColor = color for text, rgb values
;      _extra = extra  ;extra keywords for the base widget.
;
; PROCEDURE:
;       There are two modes. In autoUpDate mode, an end time and timerInterval must be
;       specified.  A non zero timerInterval triggers this mode.  At the end of the time
;       the object will destroy itself.  Of course, pressing the cancel button or closing
;       the widget also destroys the object.
;
;
;       Normal mode also has two options.  In the first you can set a default percentage
;       increment and just call the update method. Alternatively, you can call the update
;       method with a percentage.
;
;       oProgress = obj_new('krProgressBar',xsize=200,ysize=25,xoffset=scrSize[0]/2, $
;                            yoffset=scrSize[1]/2, /displayPercent)
;
;       for i=0,100 do begin
;         oProgress->upDate,i ;update is in percent
;       endfor
;
;       obj_destroy, oProgress
;
; METHODS:
;       Only a single method is meant to be called by a normal user.
;
;
;       UPDATE = Updates the progress bar. With an argument, a number between 0
;          and 100 that indicates the percent of progress bar that should be updated.
;          When called with no argument, the bar is updated with the default percentage.
; EXAMPLE:
;
;       See the example programs that accompany this file.
;
;
; MODIFICATION HISTORY:
;       Written by:  Ronn Kling, January 9, 2005
;-

;{{:|{{:|{{:|{{:|{{:|{{:|{{:|{{:|{{:|{{:|{{:|{{:|{{:|{{:|
function krstep, percentage,xsize,ysize
;creates a step mask as a function of percentage done
;this is an example for other user written mask routines

newMask = bytarr(xsize,ysize) + 255B ;start out opaque
x = bytarr(xsize*percentage > 1)+1B
y = bytarr(ysize)+1B
ramp =x#y
newMask[0,0] = ramp
return,newMask
end

;{{:|{{:|{{:|{{:|{{:|{{:|{{:|{{:|{{:|{{:|{{:|{{:|{{:|{{:|

function krLinear,percentage,xsize,ysize
;creates a linear mask as a function of percentage done
;this is an example for other user written mask routines

newMask = bytarr(xsize,ysize) + 255B ;start out opaque
x = reverse(bytscl(findgen(xsize*percentage > 1)))
y = bytarr(ysize)+1B
ramp =x#y
newMask[0,0] = ramp
return,newMask
end

;{{:|{{:|{{:|{{:|{{:|{{:|{{:|{{:|{{:|{{:|{{:|{{:|{{:|{{:|

pro krProgressBar_event,event
;event handler for the progress bar
widget_control, event.top, get_uvalue=object
method = widget_info(event.id,/uname)
call_method, method, object, event
return & end

;{{:|{{:|{{:|{{:|{{:|{{:|{{:|{{:|{{:|{{:|{{:|{{:|{{:|{{:|

pro krProgressCleanup, base
;cleanup routine
widget_control, base, get_uvalue=object
obj_destroy,object
return & end

;{{:|{{:|{{:|{{:|{{:|{{:|{{:|{{:|{{:|{{:|{{:|{{:|{{:|{{:|

function krProgressBar::createDefaultBackgroundImage,xsize,ysize
;creates a default "really cool" blue image for the background
;users can use this as an example for their own backgroundImage

;these next vectors are really color table 1 (blue-white)
red = bytscl([intarr(193),indgen(63)])
blue = bytscl([indgen(188),intarr(256-188)+188])
green = bytscl([intarr(96),indgen(256-96)])
;these next vectors are really color table 8 (green-white)
;red = bytscl([intarr(97),indgen(255-97)])
;green = bytscl([indgen(255)])
;blue = bytscl([intarr(181),indgen(256-181)])

;create a 1D array (I chose these numbers via trial and error)
;of 25 elements
image = [228,228,238,248,248,238,228,217,197,186,165,155,145,145,145, $
         145,155,166,176,186,186,186,176,166,124]
if ysize ne 25 then image=congrid(image,ysize)
image = reform(image,1,ysize)
image = rebin(image,xsize,ysize)
;If I displayed the image right now (using color table 1) it would exactly like what I want,
;but I need the RGB components!
newImage = bytarr(3,xsize,ysize)
;so here is the trick
newImage[0,*,*] = red[image]
newImage[1,*,*] = green[image]
newImage[2,*,*] = blue[image]
oImage = obj_new('IDLgrImage',newImage,interpolate=1)
return, oImage
end

;{{:|{{:|{{:|{{:|{{:|{{:|{{:|{{:|{{:|{{:|{{:|{{:|{{:|{{:|

function krProgressBar::createDefaultMaskImage,xsize,ysize
;creates a default white image for the background
;users can use this as an example for their own maskImage

image = bytarr(4,xsize,ysize) + 0B
;alpha channel is totally opaque
image[3,*,*] = 255B
oMask = obj_new('IDLgrImage',image,blend_function=[3,4])
return, oMask
end

;{{:|{{:|{{:|{{:|{{:|{{:|{{:|{{:|{{:|{{:|{{:|{{:|{{:|{{:|

pro krProgressBar::updateMask
;calls the masking function to update the mask
;also updates the percentage text if it is visible
;checks to see if the user has pressed cancel

if self.cancelFlag then begin
  obj_destroy, self
  return
endif

self.oMask->getProperty, dimension=dimension, data=data
newMask = call_function(self.maskFunction,self.currentPercentage,dimension[0],dimension[1])
data[3,*,*] = newMask
self.oMask->setProperty,data=data
if obj_valid(self.oText) then $
  self.oText->setProperty,string=string(self.currentPercentage*100,format='(i3)') + ' %'
self.oWindow->draw,self.oView
return & end

;{{:|{{:|{{:|{{:|{{:|{{:|{{:|{{:|{{:|{{:|{{:|{{:|{{:|{{:|

pro krProgressBar::drawEvents,event
;draw window events are handled here. Only expose events right now
self.oWindow->draw,self.oView
return & end

;{{:|{{:|{{:|{{:|{{:|{{:|{{:|{{:|{{:|{{:|{{:|{{:|{{:|{{:|

pro krProgressBar::cancelEvent,event
;called when the user presses cancel
self.cancelFlag = 1
return & end

;{{:|{{:|{{:|{{:|{{:|{{:|{{:|{{:|{{:|{{:|{{:|{{:|{{:|{{:|

pro krProgressBar::upDate, inPercentage, cancel=cancel
;updates the bar based upon the percentage. If percentage is greater than 1 it
;gets divided by 100.  If percentage is not present the currentPercentage is
;updated by percentage interval

if obj_valid(self.oIDLBridge) then begin
  self.oIDLBridge->setVar,'inPercentage',inPercentage 
  self.oIDLBridge->execute,"oProgress->upDate,inPercentage, cancel=cancel"
  cancel = self.oIDLBridge->getVar('cancel')
  return  
endif
cancel = self.cancelFlag

if n_params() eq 0 then begin
  self.currentPercentage = self.currentPercentage + self.percentInterval
  if self.currentPercentage ge 1.0 then self.currentPercentage=0.0 ;reset it
endif else begin
  ;make a copy to keep this from changing an incoming index
  percentage = inPercentage
  percentage=percentage/100.0
  self.currentPercentage = percentage > 0.01
endelse

if self.currentPercentage le 0.0 then return ;nonsense value
;update the mask
self->updateMask

return & end

;{{:|{{:|{{:|{{:|{{:|{{:|{{:|{{:|{{:|{{:|{{:|{{:|{{:|{{:|

pro krProgressBar::baseEvents, event
;base events, (only timer for now) are processed here

case tag_names(event,/struct) of
'WIDGET_TIMER' : begin
                 self.totalTime = self.totalTime + self.timerInterval
                 self.currentPercentage = abs(self.totalTime/self.endTime)
                 self->updateMask
                 if not obj_valid(self) then return ;user pressed cancel
                 if self.totalTime ge abs(self.endTime) then begin
                   if self.endTime gt 0.0 then begin
                     widget_control,event.top,/destroy
                     return
                   endif else begin
                     self.totalTime = 0.0
                   endelse
                 end
                 widget_control,self.base,timer=self.timerInterval
                 end
else : print,tag_names(event,/struct)
endcase

return & end

;{{:|{{:|{{:|{{:|{{:|{{:|{{:|{{:|{{:|{{:|{{:|{{:|{{:|{{:|

function krProgressBar::init,    parent, $  ; The widget ID of the group leader.
   bridge = bridge, $ ;if set then this runs in a bridge object, useful for making the cancel
                      ;button work in cpu intensive apps
   cancelButton=cancelButton, $ ; This keyword is set if a cancel button is required.
   endTime=endTime, $        ; The total time the widget should be on the display in AutoUpDate mode.
   message=message, $    ; The text of the label above the progress bar.
   title=title, $        ; The text of the top-level base title bar.
   xsize=xsize, $        ; The XSize of the progress bar in Device coordinates.
   ysize=ysize, $        ; The YSize of the progress bar in Device coordinates.
   backgroundImage = backgroundImage, $ ;background image object
   maskImage = maskImage, $ ;mask image for sliding over the background image
   maskFunction = maskFunction, $ ;function to call for the mask image opacity values
   timerInterval = timerInterval, $ ;timer interval for auto update
   percentInterval = percentInterval , $ ;default percentage increment (10%)
   displayPercent=displayPercent, $ ;if set then percentage is displayed
   textColor=textColor, $ ;color for text
   xoffset=xoffset,  $ ;offset of the base, if not present then centers
   yoffset=yoffset,  $ ;offset of the base, if not present then centers
   _extra = extra  ;extra keywords for the base

if n_elements(endTime) eq 0 then endTime=10.0
if n_elements(message) eq 0 then message='Working....'
if n_elements(title) eq 0 then title='krProgressBar'
if n_elements(xsize) eq 0 then xsize=200.0
if n_elements(ysize) eq 0 then ysize=25.0
if keyword_set(timerInterval) then self.timerInterval=timerInterval
if not keyword_set(percentInterval) then percentInterval=10.0
if not keyword_set(maskFunction) then maskFunction='krLinear'
if not keyword_set(textColor) then textColor=[255,255,255]

device,get_screen_size=scrSz
if n_elements(xoffset) eq 0 then xoffset=(scrSz[0]/2 - xsize/2)
if n_elements(yoffset) eq 0 then yoffset=(scrSz[1]/2 - ysize/2)

if keyword_set(bridge) then begin
  self.oIDLBridge = obj_new('IDL_IDLBridge')
  self.oIDLBridge->setVar,'cancelButton',keyword_set(cancelButton)
  self.oIDLBridge->setVar,'endTime',endTime
  self.oIDLBridge->setVar,'message',message
  self.oIDLBridge->setVar,'title',title
  self.oIDLBridge->setVar,'xsize',xsize
  self.oIDLBridge->setVar,'ysize',ysize
  self.oIDLBridge->setVar,'displayPercent',keyword_set(displayPercent)
  self.oIDLBridge->setVar,'tlb_frame_attr',4
  ;self.oIDLBridge->setVar,'extra',extra
  ;add more when I need them
  
  self.oIDLBridge->execute,"oProgress = obj_new('krProgressBar',cancelButton=cancelButton,message=message, tlb_frame_attr=tlb_frame_attr, displayPercent=displayPercent )"

  return,1
endif
self.endTime = endTime
self.maskFunction = maskFunction
self.percentInterval=percentInterval/100.0
;create the widgets
if n_params() eq 1 then begin
  self.base = widget_base(title=title,column=1, $
                        group_leader=parent,uname='baseEvents', $
                        floating=n_params(), xoffset=xoffset, yoffset=yoffset, $
                        _extra=extra)
endif else begin
  self.base = widget_base(title=title,column=1, $
                          uname='baseEvents', $
                          floating=n_params(), xoffset=xoffset, yoffset=yoffset, $
                          _extra=extra)
endelse                        
drawId = widget_draw(self.base, xsize=xsize, ysize=ysize, graphics_level=2, $
                     expose_events=1,uname='drawEvents',renderer=1)
if message ne '' then void = widget_label(self.base,value=message)
if keyword_set(cancelButton) then $
  void = widget_button(self.base,value='Cancel',uname='cancelEvent')

widget_control,self.base,/realize

widget_control, drawId,get_value=oWindow
self.oWindow = oWindow
self.oView = obj_new('IDLgrView',viewplane_rect=[0,0,xsize,ysize],color=[255,255,255])
oModel = obj_new('IDLgrModel')
if obj_valid(backgroundImage) eq 0 then begin
  backgroundImage = self->createDefaultBackgroundImage(xsize,ysize)
endif
if obj_valid(maskImage) eq 0 then begin
  maskImage = self->createDefaultMaskImage(xsize,ysize)
endif else begin
  ;have to double check a few things
  ;make sure it has an alpha channel and it is set to 255B
  maskImage->getProperty,data=image
  sz = size(image,/dimen)
  if sz[0] ne 4 then begin
    newImage = bytarr(4,sz[1],sz[2])
    newImage[0,*,*] = image[0,*,*]
    newImage[1,*,*] = image[1,*,*]
    newImage[2,*,*] = image[2,*,*]
    newImage[3,*,*] = 255B
    image = newImage
  endif else begin
    image[3,*,*] = 255B
  endelse
  maskImage->setProperty,data=image
  ;and you have to set the blend function
  maskImage->setProperty,blend_function=[3,4]
endelse

self.oImage = backgroundImage
self.oMask = maskImage

oModel->add,self.oImage
oModel->add,self.oMask

;check to see if we need to overlay the percentage
if keyword_set(displayPercent) then begin
  self.oText = obj_new('IDLgrText','0 %',location=[xsize/2,ysize/2,0.1], $
                        alignment=0.5, vertical_alignment=0.5 , color=textColor)
  oModel->add,self.oText
endif

self.oView->add, oModel
self.oWindow->draw,self.oView
widget_control,self.base,set_uvalue=self
if self.timerInterval ne 0.0 then begin
  no_block = 0
  just_reg = 1
  widget_control,self.base,timer=self.timerInterval
endif else begin
  no_block = 1  ;
  just_reg = 0
endelse
xmanager,'krProgressBar',self.base,just_reg=just_reg, $
         no_block=no_block, cleanup='krProgressCleanup'

return,1
end

;{{:|{{:|{{:|{{:|{{:|{{:|{{:|{{:|{{:|{{:|{{:|{{:|{{:|{{:|

pro krProgressBar::cleanup
;cleanup routine
if obj_valid(self.oImage) then obj_destroy,self.oImage
if obj_valid(self.oMask) then obj_destroy,self.oMask
if obj_valid(self.oText) then obj_destroy,self.oText
if obj_valid(self.oWindow) then obj_destroy,self.oWindow
if obj_valid(self.oView) then obj_destroy,self.oView
if obj_valid(self.oIDLBridge) then obj_destroy,self.oIDLBridge

;get rid of the base if it still exists
if widget_info(self.base,/valid) then widget_control, self.base,/destroy

return & end

;{{:|{{:|{{:|{{:|{{:|{{:|{{:|{{:|{{:|{{:|{{:|{{:|{{:|{{:|

;definition routine
pro krProgressBar__define

void = {krProgressBar, $
        oImage : obj_new(), $ ;image for the background
        oMask : obj_new(), $ ;image for the sliding mask
        oText : obj_new(), $ ;text object for the percentage
        oView : obj_new(), $ ;view object
        oWindow : obj_new(), $ ;window object
        oIDLBridge : obj_new(), $
        base : 0L, $ ;base id
        maskFunction : '', $ ;function that controls the mask opacity
        timerInterval : 0.0, $ ;timer interval for auto update
        endTime : 0.0, $ ;how long the bar stays up (seconds)
        totalTime : 0.0, $ ;how long the bar has been running
        percentInterval : 0.0 , $ ;default percentage increment
        currentPercentage : 0.0, $ ;current percentage done
        cancelFlag : 0, $ ;flag for canceling
        bounceFlag : 0} ;if set, bar bounces (not working yet)

return & end

